import argparse
from pythonosc import udp_client
import pyttsx3
from util.lrc_parse import get_lyric_timing
from util.gpt_chat import gen_chat_reply
from util.semantic_sim import semantic_sim
from util.tts_stt import text_to_speech, speech_to_text


def get_arg_parse():
    """
    Parse input arguments
    """
    parser = argparse.ArgumentParser()

    # File IO
    parser.add_argument('--lyric_dir', type=str, required=True, help='The path to directory of lrc files')
    parser.add_argument('--prompt_template_file', type=str, required=True, help='The path to the file containing the GPT-3 prompt.')

    parser.add_argument('--input_type', type=str, default="text", help='Take in input from ("text" or "mic")')

    # OSC setup
    parser.add_argument('--osc_ip', type=str, default="127.0.0.1", help='OSC message IP address')
    parser.add_argument('--osc_port', type=int, default=8338, help='OSC message port')

    # GPT-3 generation parameters
    parser.add_argument('--engine', type=str, default="text-davinci-002", help='The GPT-3 engine to use.')
    parser.add_argument('--max_tokens', type=int, default=10, required=False, help='')
    parser.add_argument('--temperature', type=float, default=0.8, required=False, help='')
    parser.add_argument('--top_p', type=float, default=0.9, required=False, help='')
    parser.add_argument('--frequency_penalty', type=float, default=0.0, required=False, help='')
    parser.add_argument('--presence_penalty', type=float, default=0.0, required=False, help='')
    parser.add_argument('--stop_tokens', nargs='+', type=str, default=None, required=False, help='Stop tokens for generation')

    args = parser.parse_args()

    return args


def send_osc(client, message):
    client.send_message("/py_to_chuck", (message.text, \
                                         int(message.start_s * 1000), \
                                         int(message.dur_s * 1000), \
                                         message.wav_file))


if __name__ == "__main__":
    args = get_arg_parse()

    # Setup OSC
    client = udp_client.SimpleUDPClient(args.osc_ip, args.osc_port)

    lyric_timing = get_lyric_timing(args.lyric_dir)

    for i in range(20):
        # Input from keyboard or microphone
        if (args.input_type == "text"):
            text = input("You: ")
        elif (args.input_type == "mic"):
            text = speech_to_text()
            print("You: ", text)
        
        # End early
        if (text.lower() == "end"):
            break

        # GPT reply
        gpt_reply = gen_chat_reply(text, args.prompt_template_file, args=args)
        engine = pyttsx3.init()
        print("ChatGPT: ", gpt_reply)
        text_to_speech(engine, gpt_reply)

        # Music reply
        lyrics = [lrc.text for lrc in lyric_timing]
        text_index = semantic_sim(gpt_reply, lyrics)
        singer_reply = lyric_timing[text_index].text
        print("SingGPT: ", singer_reply)

        print("\n")
        
        # Send osc message
        send_osc(client, lyric_timing[text_index])
    
    print("EXITING SingGPT")

    